// ****************************************************************************
#region header

// Title       : Database.cs
// Author      : Matthew Tinn
// Date        : 23/05/2013
// Description : Class for the card database

// External Sources
// http://forum.unity3d.com/threads/28500-SQLite-Class-Easier-Database-Stuff
// Rustam Ganeyev's C# Version of Javascript SQLite Code


#endregion
// ****************************************************************************
#region includes

using UnityEngine;
using System;
using System.IO;
using System.Collections;
using Mono.Data.Sqlite; // SQLite plugin

#endregion
// ****************************************************************************
public class Database
{ // Manages the execution of SQL statements which act on the database

  // **************************************************************************
  // CLASS MEMBERS

  SqliteConnection connection;   // Connection to the database
  SqliteCommand command;         // SQL command
  SqliteDataReader dataReader;   // Data reader object
        
  string sql;                    // SQL statement string

  // ****************************************************************************
  // CONSTRUCTOR
  public Database(string dbFilename)
  { // Constrcutor
    // Attempts to establish the database connection
    EstablishConnection(dbFilename);
  }

  // ****************************************************************************
  // DATABASE CONNECTION SUBPROGRAMS

  public void EstablishConnection(string dbFilename)
  { // Establishes the connection to the database
    Debug.Log("Establishing connection");

    // Creates a new SQLite database connection to the specified file
    // Uses SQLite Ver 3 & UTF-16 character encoding
    connection = new SqliteConnection("Data Source =" + dbFilename + "; Version=3; UseUTF16Encoding=True;");

    // Open the connection
    connection.Open();

    Debug.Log("Connection to " + dbFilename + " established.");
  }

  public void CloseConnection()
  { // Closes the connection to the database

    // Set the command to null
    command = null;

    // If the data reader is not null, close the connection
    // to the data reader
    if (dataReader != null)
    {
      dataReader.Close();
    }

    // Set the data reader to null
    dataReader = null;

    if (connection != null)
    { // CLose the connection if there is one
      connection.Close();
    }

    // Set the connection back to null
    connection = null;

    Debug.Log("Database disconnection logged.");
  }

  // ****************************************************************************
  // SQL SUBPROGRAMS

  public SqliteDataReader ExecuteQuery(string sqlQuery)
  { // Executes an SQL query

    // Creates the command
    command = connection.CreateCommand();

    // Store the query string as the command
    command.CommandText = sqlQuery;

    // Use the data reader to execute the command
    dataReader = command.ExecuteReader();

    return dataReader;
  }

  public SqliteDataReader ExecuteQuery(string sqlQuery, string tableName)
  { // Executes SQL Create Table queries.
    // Performs the same central function as the main subprogram
    // However stores in a text file that the table has been created
    // This prevents attempting to re-create the table by mistake
    command = connection.CreateCommand();
    command.CommandText = sqlQuery;

    dataReader = command.ExecuteReader();

    StoreTableExists(tableName);

    return dataReader;    
  }

  public SqliteDataReader CreateTable(string tableName, string[] column, string[] columnType)
  { // SQL CREATE TABLE command
      // Creates a new database table with columns as specified data types

      if (column.Length != columnType.Length)
      { // If the number of columns and number of data types specified don't match
          // throw exception
          throw new SqliteException("column.Length != columnType.Length");
      }

      // Build the SQL statement string
      string sqlQuery = "CREATE TABLE " + tableName + " (" + column[0] + " " + columnType[0];

      for (int i = 1; i < column.Length; i++)
      {
          sqlQuery += ", " + column[i] + " " + columnType[i];
      }

      sqlQuery += ")";

      Debug.Log(sqlQuery);

      // Execute the statement using the create table version
      // of ExecuteQuery(...)
      return ExecuteQuery(sqlQuery, tableName);
  }

  public void StoreTableExists(string tableName)
  { // Stores if a table exists to prevent erroneous
      // re-creating of tables

      // Create an instance of the FileIO class to handle IO
      FileIO lineWriter = new FileIO();

      // Write the table name to the table check file
      lineWriter.WriteToFile("tableCheck.txt", tableName);

      Debug.Log("Table stored as created.");
  }

  public bool CheckTableExists(string tableName)
  { // Checks if a table in a database exists

      // Max number of tables can handle
      const int MAX_NUM_TABLES = 200;

      // Get the path of the table check file
      string path;
      path = System.IO.Directory.GetCurrentDirectory();
      path += "\tableCheck.txt";

      // Create an array to hold table names
      string[] tableNameArray = new string[MAX_NUM_TABLES];

      // Create an instance of the custom FileIO class to read the file
      FileIO lineReader = new FileIO();

      // Read from the file, using the return boolean to pick up blank or
      // non-existing file errors
      // This will return by reference an array of table names stored in the file
      bool fileCheck = lineReader.ReadFromFile("tableCheck.txt", ref tableNameArray, MAX_NUM_TABLES);

      bool tableExists = false; // Boolean indicates whether table exisits in DB

      if (!fileCheck)
      { // If no file or blank file, output debug message
          Debug.Log("Error: Non-existant file or blank file.");
      }
      else
      { // Loop through the retrieved table names
          // Existance is shown by a match between the table passed in
          // and one of the retrieved tables
          for (int i = 0; i < MAX_NUM_TABLES; i++)
          {
              if (String.Equals(tableName, tableNameArray[i]))
              {
                  tableExists = true;
              }
          }
      }

      return tableExists;
  }

  public SqliteDataReader DeleteAll(string tableName)
  { // Deletes all tuples from the table
      string sqlQuery = "DELETE FROM " + tableName;
      return ExecuteQuery(sqlQuery);
  }


  public SqliteDataReader SelectAll(string tableName)
  { // SQL SELECT ALL command
    string sqlQuery = "SELECT * FROM " + tableName;
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader SelectFrom(string tableName, string[] items)
  { // SQL SELECT FROM command

    // Build the SQL query
    string sqlQuery = "SELECT " + items[0];

    for (int i = 1; i < items.Length; i++)
    {
      sqlQuery += ", " + items[i];
    }

    sqlQuery += " FROM " + tableName;

    // Execute the SQL query
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader SelectFromWhere(string tableName, string[] items, string[] column, string[] operation, string[] values, bool numeric)
  { // SQL SELECT FROM WHERE command
    if (column.Length != operation.Length || operation.Length != values.Length)
    { // Throw an exception if the number of columns != number of operations
      // or number of operations != number of values
      throw new SqliteException("column.Length != operation.Length || operation.Length != values.Length");
    }
    Debug.Log("In SelectfromWhere");
    // Build the SQL query
    string sqlQuery = "SELECT " + items[0];

    for (int i = 1; i < items.Length; i++)
    {
      sqlQuery += ", " + items[i];
    }

    sqlQuery += " FROM " + tableName + " WHERE " + column[0] + operation[0];

    if (!numeric)
    {
      sqlQuery += "'";
    }

    sqlQuery += values[0];

    if (!numeric)
    {
      sqlQuery += "'";
    }

    Debug.Log(sqlQuery);
    Debug.Log("Out SelectFromWhere");
    // Execute the SQL query
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader SelectFromWhereAnd(string tableName, string[] items, string[] column, string[] operation, string[] values)
  { // SQL SELECT FROM WHERE AND command

    if (column.Length != operation.Length || operation.Length != values.Length)
    { // Throw an exception if the number of columns != number of operations
      // or number of operations != number of values
      throw new SqliteException("column.Length != operation.Length || operation.Length != values.Length");
    }

    // Build the SQL query string
    string sqlQuery = "SELECT " + items[0];

    for (int i = 1; i < items.Length; i++)
    {
      sqlQuery += ", " + items[i];
    }

    sqlQuery += " FROM " + tableName + " WHERE " + column[0] + operation[0] + "'" + values[0] + "' ";

    for (int i = 1; i < column.Length; i++)
    {
      sqlQuery += " AND " + column[i] + operation[i] + "'" + values[i] + "' ";
    }

    // Execute the SQL query
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader InsertInto (string tableName, string[] values)
  { // SQL INSERT command
    
    // Build the SQL command for inserting data into a table
    string sqlQuery = "INSERT INTO " + tableName + " VALUES (" + values[0];

    for (int i = 1; i < values.Length; i++)
    {
      sqlQuery += ", " + values[i];
    }

    sqlQuery += ")";

    Debug.Log(sqlQuery);

    // Execute the statement
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader InsertIntoSpecific (string tableName, string[] columns, string[] values)
  { // SQL INSERT INTO COLUMNS command

    // Build the SQL command for inserting data into a table on specified columns
    
    if (columns.Length != values.Length)
    { // Throws an exception if the number of values does not match
      // the number of columns
      throw new SqliteException("columns.Length != values.Length");
    }

    // Build the insert statement
    string sqlQuery = "INSERT INTO " + tableName + "(" + columns[0];

    for (int i = 1; i < columns.Length; i++)
    {
      sqlQuery += ", " + columns[i];
    }

    sqlQuery += ") VALUES (" + values[0];

    for (int i = 1; i < values.Length; i++)
    {
        sqlQuery += ", " + values[i];
    }

    sqlQuery += ")";

    // Should be
    // INSERT INTO table(x, y, z) VALUES (a, b, c)
    Debug.Log(sqlQuery);
    
    // Execute the SQL statement
    return ExecuteQuery(sqlQuery);
  }

  
  public SqliteDataReader UpdateWhereOnce(string tableName, string[] updtColumns,  string[] updtOper, string updtFunction,
                                      string[] condColumns, string[] condOpers, string[] condValues)
  { // Performs an UPDATE with a WHERE clause on one piece of data

    // Build the SQL query
    string sqlQuery = "UPDATE " + tableName + " SET " + updtColumns[0] +
                       updtOper[0] + updtFunction;

    for (int i = 1; i < updtColumns.Length; i++)
    {
      sqlQuery += ", " + updtColumns[i];
      sqlQuery += " " + updtOper[i];
      sqlQuery += " " + updtFunction;
    }

    sqlQuery += " WHERE " + condColumns[0] + " " + condOpers[0] + " " + condValues[0];

    sqlQuery += ";";

    // Ouput the query to the console
    Debug.Log(sqlQuery);

    // Execute the query
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader Commit()
  { // Commits chnages to the database
    string sqlQuery = "COMMIT;";
    return ExecuteQuery(sqlQuery);
  }

  public SqliteDataReader Rollback()
  { // Rolls back changes to the database
    string sqlQuery = "ROLLBACK;";
    return ExecuteQuery(sqlQuery);
  }

  
} // End of class
