﻿// ****************************************************************************
#region header

// Title       : MatchMeGame.cs
// Author      : Matthew Tinn
// Date        : 23/05/2013
// Description : Codebase for the Match Me game


#endregion
// ****************************************************************************
#region includes

using UnityEngine;
using System;
using System.IO;
using System.Collections;

#endregion
// ****************************************************************************
public class MatchMeGame : MonoBehaviour 
{
  // ****************************************************************************
  // CLASS MEMBERS

  // Grid
  private int columns;                   // The number of columns in a card grid, typically 1.
  private int rows;                      // The number of rows in a card grid, typically 5.

  private const int cardWidth = 100;     // The physical width of a card on the screen
  private const int cardHeight = 150;    // The physical height of a card on the screen

  private const int MAX_DECK_CARDS = 25; // The maxiumum number of cards in the deck
  private const int MAX_HAND_CARDS = 5;  // The maximum number of cards in play, known as a hand
  private const int MAX_SEL_CARDS = 2;   // The maximum number of cards that can be selected
  private int selectedCardCount;         // The count of cards selected in a turn
  private int selectedCardCountLeft;     // The count of cards selected from left column
  private int selectedCardCountRight;    // The count of cards selected from right column
  private int prevCardIndLeft;           // The previous card selected in the left column
  private int prevCardIndRight;          // The previosu card selected in the right column

  // Cards
  private DeckManager deckMan;           // Retrieves cards from the database
  private Card[] deck;                   // The selection of cards to be reviewed
  private Card[] hand;                   // The cards currently in play
  private Card[] selectedCards;          // The two cards selected

  // Arrays that hold the indexes of cards for shuffling
  private int[] frontIndexArray = new int [5] {0, 1, 2, 3, 4};
  private int[] backIndexArray = new int [5] {0, 1, 2, 3, 4};
  
  private Card[] erasedCards;            // The array of cards erased    
  private int numErased;                 // The number of cards erased

  // Game
  private double matchesNeededToWin;     // Number of matches that need to be made to win, typically 5
  private int matchesMade;               // Count of the number of matches made
  private bool completeRound;            // Flag for a completed round
  private bool completeGame;             // Flag for completed game
  private bool playAllowed;              // Allows the suspension of play
  private int roundsCompleted;           // The number of rounds completed

  // UI

  private const int whiteSpace = 100;    // Holder for white space between UI groups
  private const int areaTop = 10;        // Holder for the top indentation for UI groups

  private int handStart;                 // Stores the starting index of the hand

  private int frameCount;                // Frame count timer for play control

  // Random Numbers
  // For shuffling
  int rand1;
  int rand2;

  // Buttons

  private Rect quitButton;               // Rect for the quit button

  public GUISkin matchMeSkin;            // GUISkin for UI

  // ****************************************************************************
  // CONSTRUCTOR

  public MatchMeGame()
  { // Initialise grid variables
    InitGame();
  }

  // ****************************************************************************
  // INITIALISATION

  public void InitGame()
  {
    // Initialise the columns and rows of the grid.
    columns = 0;
    rows = 0;

    // Initialise game variables
    matchesNeededToWin = 0;
    matchesMade = 0;
    completeRound = false;
    completeGame = false;
    playAllowed = true;
    roundsCompleted = 0;
    frameCount = 0;
    numErased = 0;

    selectedCardCountLeft = 0;
    selectedCardCountRight = 0;
    prevCardIndLeft = 0;
    prevCardIndRight = 0;
  }



  // ****************************************************************************
  // DECK BUILDING

  public void BuildDeck()
  { // Builds the deck of cards

      // Create a new deck manager object
      deckMan = new DeckManager();

      // Create a new deck of cards
      deck = new Card[MAX_DECK_CARDS];

      // Initialise the deck of cards
      for (int i = 0; i < MAX_DECK_CARDS; i++)
      {
          deck[i] = new Card();
      }

      // Initialise the hand to play with
      hand = new Card[MAX_HAND_CARDS];

      for (int i = 0; i < MAX_HAND_CARDS; i++)
      {
          hand[i] = new Card();
      }

      // Initialise the selected cards array
      selectedCards = new Card[MAX_SEL_CARDS];

      for (int i = 0; i < MAX_SEL_CARDS; i++)
      {
          selectedCards[i] = new Card();
      }

      // Initialise the erased cards array
      erasedCards = new Card[MAX_HAND_CARDS];

      for (int i = 0; i < MAX_HAND_CARDS; i++)
      {
          erasedCards[i] = new Card();
      }

      // Call DeselectCards to makes sure there are no selected cards
      DeselectCards();

      // Retrieve the card data from the database
      deckMan.NewRetrieveCardData(ref deck, MAX_DECK_CARDS); // VSHOW
  }

  //-----------------------------------//

  public void BuildHand()
  { // Builds a hand of cards

    // Create a new hand of cards
    hand = new Card[MAX_HAND_CARDS];

    // Initialise the hand of cards
    for (int i = 0; i < MAX_HAND_CARDS; i++)
    {
      hand[i] = new Card();
    }

    // Pick cards out of the deck and add them to the hand
    // Dependent on the number of rounds completed
    // Takes the next five cards each time a new round begins

    if (roundsCompleted == 1)
    {
      // Take cards 5 - 9
      handStart = 5;
    }
    else if (roundsCompleted == 2)
    {
      // Take cards 10 - 14
      handStart = 10;
    }
    else if (roundsCompleted == 3)
    {
      // Take cards 15 - 19
      handStart = 15;
    }
    else if (roundsCompleted == 4)
    {
      // Take cards 20 - 24
      handStart = 20;
    }
    else // If roundsCompleted == 0 or any other number
    {
      // Take cards 0 - 4
      handStart = 0;
    }

    // Set the cards in the hand
    SetHandCards(handStart);

  }

  //-----------------------------------//

  public void SetHandCards(int deckCard)
  { // Contains a re-usable loop for setting the cards
    // in the hand, taking a starting number for
    // the deck as a parameter

    for (int i = 0; i < MAX_HAND_CARDS; i++)
    { // Take 5 cards out of the deck and put in hand
      hand[i] = deck[deckCard];

      deckCard++;
    }
  }

  // ****************************************************************************
  // CARD SELECTION

  public void SelectCards(int cardIndex, int sideClicked) // VSHOW
  { // Selects a card, placing it in the centre column
    // When the user clicks on a card face

    if (playAllowed)
    { // If the user can play
      if (sideClicked == 1 && selectedCardCountLeft == 0)
      { // If the user clicks the left column and there is not 
        // currently a card selected from it

        // Set that card as the selected card and increment the click count
        hand[cardIndex].setSelected();
        hand[cardIndex].IncrementClickCount();

        // Increment the overall number of cards selected and the number of
        // cards selected from the left column
        selectedCardCount++;
        selectedCardCountLeft++;

        // Store this card as the previously selected card
        // for if the user selects another card from the same column
        prevCardIndLeft = cardIndex;

        // Set the selected card in position 0 to the current card
        // in the hand
        selectedCards[0].SetCardID(hand[cardIndex].GetCardID());
        selectedCards[0].SetBoxID(hand[cardIndex].GetBoxID());
        selectedCards[0].setFront(hand[cardIndex].GetFront());
        selectedCards[0].setBack(hand[cardIndex].GetBack());
      }
      else if (sideClicked == 1 && selectedCardCountLeft > 0)
      { // If a card is clicked in a column with a selected card

        // Deselect the current card
        hand[prevCardIndLeft].setUnselected();

        // Set the card just selected as the selected card
        hand[cardIndex].setSelected();
        prevCardIndLeft = cardIndex;

        // Set the selected card in position 0 to the current card
        // in the hand
        selectedCards[0].SetCardID(hand[cardIndex].GetCardID());
        selectedCards[0].SetBoxID(hand[cardIndex].GetBoxID());
        selectedCards[0].setFront(hand[cardIndex].GetFront());
        selectedCards[0].setBack(hand[cardIndex].GetBack());
      }
      else if (sideClicked == 2 && selectedCardCountRight == 0)
      {
        // If user clicks right column and no cards in that column have been
        // selected, set cards as selected and increment click count
        hand[cardIndex].setSelected();
        hand[cardIndex].IncrementClickCount();

        // Increament selected card counts
        selectedCardCount++;
        selectedCardCountRight++;

        // Set the selected card in position 1 to the current card
        // in the hand
        selectedCards[1].SetCardID(hand[cardIndex].GetCardID());
        selectedCards[1].SetBoxID(hand[cardIndex].GetBoxID());
        selectedCards[1].setFront(hand[cardIndex].GetFront());
        selectedCards[1].setBack(hand[cardIndex].GetBack());
      }
      else if (sideClicked == 2 && selectedCardCountRight > 0)
      { // If a card in the right column is selected and card in that column
        // has been selected, deselect currently stored card
        hand[prevCardIndRight].setUnselected();
        
        // Set the newly selected card as the selected card
        hand[cardIndex].setSelected();
        prevCardIndRight = cardIndex;

        // Set the selected card in position 1 to the current card
        // in the hand
        selectedCards[1].SetCardID(hand[cardIndex].GetCardID());
        selectedCards[1].SetBoxID(hand[cardIndex].GetBoxID());
        selectedCards[1].setFront(hand[cardIndex].GetFront());
        selectedCards[1].setBack(hand[cardIndex].GetBack());
      }

      if (selectedCardCount == 2)
      { // When the player has selected 2 cards, stop them from selecting more
        // and allow OnGUI() to delay then check for match
        playAllowed = false;
      }
    }
  }
  //-----------------------------------//
  public void DeselectCards()
  { // De-selects the current cards
    
    // Match the selected card ID with the hand card ID
    // and deselect the card in the hand
    for (int i = 0; i < MAX_HAND_CARDS; i++)
    {
      for (int j = 0; j< MAX_SEL_CARDS; j++)
      {
        if (selectedCards[j].GetCardID() == hand[i].GetCardID())
        {
          Debug.Log("Deselect cards: " + i);
          hand[i].setUnselected();
        }
      }
    }

    for (int i = 0; i < MAX_SEL_CARDS; i++)
    { // Reset the the selected cards array
        selectedCards[i].SetCardID(-1);
        selectedCards[i].SetBoxID(-1);
        selectedCards[i].setFront("");
        selectedCards[i].setBack("");
    }

    // Reset the selected card count
    selectedCardCount = 0;
    selectedCardCountLeft = 0;
    selectedCardCountRight = 0;

    // Allow play
    playAllowed = true;

  }

  // ****************************************************************************
  // CARD MATCHING

  public bool CheckMatch() // VSHOW
  { // Checks a card to see if it has been matched

    // Output debug messages to the console
    DebugGetSelected();

    if (selectedCards[0].GetCardID() == selectedCards[1].GetCardID())
    { // If the cardIDs of the front and back match, match is true
      Debug.Log(selectedCards[0].GetCardID() + " is selected.");
      selectedCards[0].SetMatched();
      selectedCards[1].SetMatched();

      for (int i = 0; i < MAX_HAND_CARDS; i++)
      { // Set matched cards in the hand to matched
        if (hand[i].GetSelected())
        {
          hand[i].SetMatched();

          // Add the hand card to the erased cards array
          erasedCards[numErased].SetCardID(hand[i].GetCardID());
          erasedCards[numErased].SetBoxID(hand[i].GetBoxID());
          erasedCards[numErased].setFront(hand[i].GetFront());
          erasedCards[numErased].setBack(hand[i].GetBack());
        }
      }

      // Increment number of cards erased
      numErased++;

      // Increment matches made
      matchesMade++;

      Debug.Log("Matches made: " + matchesMade + " and matches to win "+ matchesNeededToWin);

      if (matchesMade >= matchesNeededToWin)
      { // If the matches made equals matches needed to win
          // Player has won the round
          completeRound = true;
      }

      // Deselect the current cards
      DeselectCards();

      // Re-activate play
      playAllowed = true;

      return true;
    }
    else
    { // If the selected cards don't match, deselect and player can choose
      // new cards
      DeselectCards();
      playAllowed = true;

      return false;
    }
  }

  // ****************************************************************************
  // GAME START & LOAD SUBPROGRAMS

  void Start()
  {
    // Set the size of a grid
    columns = 1;
    rows = 5;

    // Set up the game variables
    matchesNeededToWin = MAX_DECK_CARDS / rows;
    playAllowed = true;

    // Set position of quit button
    this.quitButton.Set(10, 10, 100, 50);

    // Set GUISkin
    matchMeSkin = Resources.Load("MatchMeSkin") as GUISkin;

    // Build the deck of cards
    BuildDeck();

    // Create a new round of play
    this.roundsCompleted = 0;
    CreateFirstRound();

    Debug.Log("Start: Out");
  }

  void Reload()
  {
    // Set the size of a grid
    columns = 1;
    rows = 5;

    // Set up the game variables
    matchesNeededToWin = MAX_DECK_CARDS / rows;
    playAllowed = true;

    this.quitButton.Set(10, 10, 100, 50);

    matchMeSkin = Resources.Load("MatchMeSkin") as GUISkin;

    BuildHand();

    Debug.Log("Start: Out");
  }

  //-----------------------------------//

  void CreateFirstRound()
  { // Creates the first round of the gane
    // Perform start of game initialisation
      BuildHand();
      SetRandomNumbers();
      matchesMade = 0;
      completeRound = false;
      playAllowed = true;
  }

  void CreateNewRound()
  { // Creates a new round
    // Re-builds the hand of cards
    // and resets appropraite variables
    InitGame();
    Reload();
  }

  //-----------------------------------//

  // ****************************************************************************
  // GUI SUBRPOGRAMS

  void OnGUI()
  {
    // Set the GUI skin
    GUI.skin = this.matchMeSkin;

    // Use round control
    if (completeRound)
    {
      // Update cards
      deckMan.UpdateCardData(hand, MAX_HAND_CARDS, "character");
      
      // Create completed round prompt
      CreateRoundPrompt();
    }
    else if (!completeRound && !completeGame)
    {
      // Draw the grid
      GridBuilder();

      if (!playAllowed)
      { // Manage checking for matched cards
        frameCount++;

        if (frameCount >= 50)
        {
          CheckMatch();
          frameCount = 0; // VSHOWBUG
        }
      }

      if (GUI.Button(quitButton, "Quit"))
      { // Manage pressing of quit button
        Debug.Log("User has quit.");
        Application.LoadLevel("MainMenu");
      }
    }
    else if (completeGame)
    {
    }
  } // End OnGUI

  public void GridBuilder()
  { // This builds the main UI

      // Set the starting position of the cards
      float cardTop = 10;
      float cardLeft = 120;

      // Loop for placing the left 5 cards
      for (int i = 0; i < 5; i++)
      {
          if (this.hand[i].GetMatched() == false)
          {
              if (GUI.Button(new Rect(cardLeft, cardTop, cardWidth, cardHeight), hand[frontIndexArray[i]].GetFront()))
              {
                  // Set as a selected card
                  this.SelectCards(frontIndexArray[i], 1);
              }
          }

          // Increase the card top to space out the cards
          cardTop += cardHeight + 10;

          if (!playAllowed)
          {
              if (frameCount > 50)
              {
                  frameCount = 0;
              }
          }
      } // end for

      cardTop = 10;

      // loop for placing the right hand cards
      for (int k = 0; k < 5; k++)
      {
          if (hand[backIndexArray[k]].GetMatched() == false)
          {
              if (GUI.Button(new Rect(Screen.width - (cardLeft * 2), cardTop, cardWidth, cardHeight), hand[backIndexArray[k]].GetBack()))
              {
                  // Set as a selected card
                  this.SelectCards(backIndexArray[k], 2);
              } // end if
          }

          cardTop += cardHeight + 10;
      } // end for


      // Place the selected card area
      cardTop = areaTop;

      float selectCardLeft = (Screen.width / 2) - cardWidth;

      GUI.Button(new Rect(selectCardLeft, cardTop, cardWidth, cardHeight), selectedCards[0].GetFront());

      cardTop += cardHeight + 10;

      GUI.Button(new Rect(selectCardLeft, cardTop, cardWidth, cardHeight), selectedCards[1].GetBack());
  }

  //-----------------------------------//

  void CreateRoundPrompt()
  { // Creates a UI prompt when the user completes a round

    // Define the size of the prompt box
    int promptWidth = 400;
    int promptHeight = 400;

    // Begin the UI group
    GUI.BeginGroup(new Rect(Screen.width/2 - promptWidth/2,
                            Screen.height/2 - promptHeight/2,
                            promptWidth, promptHeight));

    // Create a box congratualting the player on completing the round
    GUI.Box(new Rect(0, 0, promptWidth, promptHeight/8), "Congratulations! Round complete!");

    // Create a button to enable the user to play another round
    if (GUI.Button(new Rect(promptWidth/2 - promptWidth/8, promptHeight / 4 + 10, promptWidth / 4, promptHeight / 8), "Back to Menu"))
    {
      // Create new round
      //CreateNewRound();
      Application.LoadLevel("MainMenu");
    }

    GUI.EndGroup();

  }

  // ****************************************************************************
  // UTILITIES

  private void SetRandomNumbers()
  {
    rand1 = UnityEngine.Random.Range(0, 2);
    rand2 = UnityEngine.Random.Range(0, 2);
  }

  // ****************************************************************************
  // DEBUG

  private void DebugGetMatched()
  { // Outputs to console what getMatched returns for a card
    for (int i = 0; i < MAX_HAND_CARDS; i++)
    {
      Debug.Log("Card " + i + ".getmatched returns " + hand[i].GetMatched().ToString());
    }
  }

  private void DebugGetSelected()
  { // Outputs to console what getSelected returns for a card
    for (int i = 0; i < MAX_HAND_CARDS; i++)
    {
      Debug.Log("Card " + i + ".getselected returns " + hand[i].GetSelected().ToString());
    }
  }
}
